"""
- Make the base class first. Set COLLECTION explicitly - it is not
  generated automatically from the class name but inherited instead.
- Subclass your base class to define submodels. Set SUBMODEL to a string
  that will identify your submodel in the DB.
- You can further subclass your submodels. To avoid saving such abstract
  intermediate models do not set SUBMODEL.
- Register the submodel with the base model

It is possible to register an arbitrary function instead of a proper class.
It may be particularly useful if the correct class depends on something
other than `submodel` field. The function will get **data from the DB and
should return an model object.

If you decide to do it you will likely have to override _preprocess_query()
on your submodels to keep the expected find/destroy/update behaviour

"""
from .fields import StringField
from .storable_model import StorableModel
from .sharded_model import ShardedModel
from ..errors import IntegrityError, MissingSubmodel, WrongSubmodel, UnknownSubmodel


class BaseSubmodelMixin:

    SUBMODEL = None

    submodel: StringField(required=True)

    __submodel_loaders__ = None

    def __init__(self, attrs, **kwargs):
        super().__init__(attrs, **kwargs)
        if self.is_new:
            if not self.SUBMODEL:
                raise IntegrityError(
                    f"Attempted to create an object of abstract model {self.__class__.__name__}")
            if "submodel" in attrs:
                raise IntegrityError(
                    "Attempt to override submodel for a new object")
            self.submodel = self.SUBMODEL
        else:
            if not self.submodel:
                raise MissingSubmodel(f"{self.__class__.__name__} has no submodel in the DB. Bug?")
            self._check_submodel()

    def _check_submodel(self):
        if self.submodel != self.SUBMODEL:
            raise WrongSubmodel(f"Attempted to load {self.submodel} as"
                                f" {self.__class__.__name__}. Correct submodel"
                                f" would be {self.SUBMODEL}. Bug?")

    def validate(self):
        super().validate()
        self._check_submodel()

    @classmethod
    def register_submodel(cls, name, ctor):
        if cls.SUBMODEL:
            raise IntegrityError("Attempted to register submodel with another submodel")
        if not cls.__submodel_loaders__:
            cls.__submodel_loaders__ = {}
        if name in cls.__submodel_loaders__:
            raise IntegrityError(f"Submodel {name} already registered")
        cls.__submodel_loaders__[name] = ctor

    @classmethod
    def _preprocess_query(cls, query):
        if not cls.SUBMODEL:
            return query
        return {
            "submodel": cls.SUBMODEL,
            **query
        }

    @classmethod
    def _ctor(cls, attrs, **kwargs):
        if "submodel" not in attrs:
            raise MissingSubmodel(f"{cls.__name__} has no submodel in the DB. Bug?")
        if not cls.__submodel_loaders__:
            return cls(attrs, **kwargs)
        submodel_name = attrs["submodel"]
        if submodel_name not in cls.__submodel_loaders__:
            raise UnknownSubmodel(f"Submodel {submodel_name} is not registered with {cls.__name__}")
        return cls.__submodel_loaders__[submodel_name](attrs, **kwargs)


class StorableSubmodel(BaseSubmodelMixin, StorableModel):
    pass


class ShardedSubmodel(BaseSubmodelMixin, ShardedModel):
    pass



import unittest
import logging
import inspect
import sys
from types import ModuleType
from . import Command

MODULE_NAME = "_all_tests"


class Test(Command):

    def run(self):
        logging.disable(logging.CRITICAL)
        all_tests_module = ModuleType(MODULE_NAME)

        glasskit_test_mod = __import__("glasskit.tests").tests
        app_test_mod = __import__(f"{self.app.app_name}.tests").tests

        for module in [glasskit_test_mod, app_test_mod]:
            for objname in dir(module):
                obj = getattr(module, objname)
                if inspect.isclass(obj) and issubclass(obj, unittest.TestCase) and obj != unittest.TestCase:
                    setattr(all_tests_module, objname, obj)

        sys.modules[MODULE_NAME] = all_tests_module
        unittest.main(argv=["glass.py test", "-b"], module=MODULE_NAME)



from . import Command
from glasskit.uorm.models.base_model import BaseModel
from glasskit.uorm.models.storable_model import StorableModel
from glasskit.uorm.models.sharded_model import ShardedModel
from glasskit.uorm.models.submodel import ShardedSubmodel, StorableSubmodel
from glasskit import ctx
import os
import inspect
import importlib


class Index(Command):
    def find_model_classes(self):
        classes = []
        model_dir = os.path.join(self.app.app_dir, "models")
        for filename in os.listdir(model_dir):
            if filename.endswith(".py"):
                mod_name = os.path.basename(filename)[:-3]
                mod_name = f"{self.app.app_name}.models.{mod_name}"
                module = importlib.import_module(mod_name)
                for key in dir(module):
                    obj = getattr(module, key)
                    if inspect.isclass(obj) and issubclass(obj, BaseModel):
                        if obj not in [
                            BaseModel,
                            StorableSubmodel,
                            ShardedSubmodel,
                            StorableModel,
                            ShardedModel,
                        ]:
                            classes.append(obj)
        return classes

    def run(self):
        model_classes = self.find_model_classes()
        for cls in model_classes:
            ctx.log.info("generating indexes for model %s", cls.__name__)
            cls.ensure_indexes()



from unittest import TestCase
from ..uorm.db import db


class MongoMockTest(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        db.configure({
            "meta": {"uri": "mongodb://fake_server:27017/unittest_meta"},
            "shards": {
                "s1": {"uri": "mongodb://fake_server:27017/unittest_s1"},
                "s2": {"uri": "mongodb://fake_server:27017/unittest_s2"},
                "s3": {"uri": "mongodb://fake_server:27017/unittest_s3"},
                "s4": {"uri": "mongodb://fake_server:27017/unittest_s4"},
            }
        }, mock=True)



import itertools
from bson import ObjectId
from .mongo_mock_test import MongoMockTest
from ..uorm.db import db
from ..uorm.models.fields import Field
from ..uorm.models.submodel import StorableSubmodel, ShardedSubmodel
from ..uorm.errors import WrongSubmodel, MissingSubmodel, IntegrityError


class TestStorableSubmodel(MongoMockTest):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        class TestBaseModel(StorableSubmodel):
            COLLECTION = "test"
            field1: Field()
            field2: Field()

        class Submodel1(TestBaseModel):

            SUBMODEL = "submodel1"
            field3: Field()

        class Submodel2(TestBaseModel):

            SUBMODEL = "submodel2"
            field4: Field()

        TestBaseModel.register_submodel(Submodel1.SUBMODEL, Submodel1)
        TestBaseModel.register_submodel(Submodel2.SUBMODEL, Submodel2)

        cls.base_model = TestBaseModel
        cls.submodel1 = Submodel1
        cls.submodel2 = Submodel2

    def setUp(self) -> None:
        super().setUp()
        self.base_model.destroy_all()

    def test_wrong_input(self):
        with self.assertRaises(WrongSubmodel):
            self.submodel1({"_id": ObjectId(), "field1": "value", "submodel": "wrong"})
        with self.assertRaises(MissingSubmodel):
            self.submodel1({"_id": ObjectId(), "field1": "value"})
        with self.assertRaises(IntegrityError):
            self.submodel1({"field1": "value", "submodel": "my_submodel"})
        with self.assertRaises(WrongSubmodel):
            obj = self.submodel1({"field1": "value"})
            obj.submodel = "wrong"
            obj.save()

    def test_submodel_field(self):
        obj = self.submodel1({})
        self.assertTrue(hasattr(obj, "submodel"))
        self.assertEqual(obj.submodel, self.submodel1.SUBMODEL)
        obj.save()
        obj.reload()
        self.assertEqual(obj.submodel, self.submodel1.SUBMODEL)
        db_obj = self.submodel1.get(obj._id)
        self.assertEqual(db_obj.submodel, self.submodel1.SUBMODEL)

    def test_inheritance(self):

        class Submodel1(self.base_model):
            SUBMODEL = "submodel1"

        class Submodel1_1(Submodel1):
            pass

        self.assertEqual(self.base_model.collection, Submodel1.collection)
        self.assertEqual(Submodel1.collection, Submodel1_1.collection)
        self.assertEqual(Submodel1.SUBMODEL, Submodel1_1.SUBMODEL)

    def test_abstract(self):
        with self.assertRaises(IntegrityError):
            self.base_model({})

        with self.assertRaises(IntegrityError):
            class C(self.base_model):
                pass  # no SUBMODEL
            C({})

        with self.assertRaises(IntegrityError):
            class C(self.submodel1):
                SUBMODEL = "c"
            self.submodel1.register_submodel("c", C)

    def test_fields_inheritance(self):
        m1 = self.submodel1({})
        m2 = self.submodel2({})

        self.assertCountEqual(m1.__fields__, ["_id", "submodel", "field1", "field2", "field3"])
        self.assertCountEqual(m2.__fields__, ["_id", "submodel", "field1", "field2", "field4"])

    def _create_objs(self):
        """Returns two lists of objects. Objects in the same positions only differ in their submodel"""
        values = [1, 2, 3]
        objs1 = [self.submodel1({"field1": v, "field2": v}) for v in values]
        objs2 = [self.submodel2({"field1": v, "field2": v}) for v in values]
        for obj in itertools.chain(objs1, objs2):
            obj.save()
        return objs1, objs2

    def test_isolation_find(self):
        objs1, objs2 = self._create_objs()
        self.assertCountEqual(
            self.submodel1.find().all(),
            objs1,
        )
        self.assertCountEqual(
            self.submodel2.find().all(),
            objs2,
        )
        self.assertCountEqual(
            self.base_model.find().all(),
            objs1 + objs2,
        )

        self.assertCountEqual(
            self.submodel1.find({"field1": objs1[0].field1}).all(),
            [objs1[0]],
        )
        self.assertCountEqual(
            self.base_model.find({"field1": objs1[0].field1}).all(),
            [objs1[0], objs2[0]],
        )

    def test_isolation_update(self):
        objs1, objs2 = self._create_objs()
        obj1 = objs2[0]
        obj1.field2 = "new_value"
        self.submodel2.update_many(
            {"field1": obj1.field1},
            {"$set": {"field2": obj1.field2}}
        )
        self.assertCountEqual(
            self.base_model.find({"field2": obj1.field2}),
            [obj1]
        )

        obj1 = objs1[1]
        obj2 = objs2[1]
        obj1.field2 = "newer_value"
        obj2.field2 = "newer_value"
        self.base_model.update_many(
            {"field1": obj1.field1},
            {"$set": {"field2": obj1.field2}}
        )
        self.assertCountEqual(
            self.base_model.find({"field2": obj1.field2}),
            [obj1, obj2]
        )

    def test_isolation_destroy(self):
        objs1, objs2 = self._create_objs()
        to_destroy = objs2.pop()
        to_keep = objs1[-1]
        self.submodel2.destroy_many({"field1": to_destroy.field1})
        self.assertListEqual(
            self.submodel2.find({
                "field1": to_destroy.field1
            }).all(),
            []
        )
        self.assertListEqual(
            self.base_model.find({
                "field1": to_destroy.field1
            }).all(),
            [to_keep]
        )

        to_destroy = objs1[0]
        objs1 = objs1[1:]
        objs2 = objs2[1:]
        self.base_model.destroy_many({"field1": to_destroy.field1})
        self.assertCountEqual(
            self.base_model.find().all(),
            objs1 + objs2,
        )

    def test_double_register(self):
        with self.assertRaises(IntegrityError):
            class NewSubmodel(self.base_model):
                SUBMODEL = self.submodel2.SUBMODEL
            self.base_model.register_submodel(
                NewSubmodel.SUBMODEL, NewSubmodel)


class TestShardedSubmodel(MongoMockTest):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        class TestBaseModel(ShardedSubmodel):
            COLLECTION = "test"
            field1: Field()
            field2: Field()

        class Submodel1(TestBaseModel):

            SUBMODEL = "submodel1"

        class Submodel2(TestBaseModel):

            SUBMODEL = "submodel2"

        TestBaseModel.register_submodel(Submodel1.SUBMODEL, Submodel1)
        TestBaseModel.register_submodel(Submodel2.SUBMODEL, Submodel2)

        cls.base_model = TestBaseModel
        cls.submodel1 = Submodel1
        cls.submodel2 = Submodel2

    def setUp(self) -> None:
        super().setUp()
        for shard in db.shards:
            self.base_model.destroy_all(shard)

    def test_wrong_input(self):
        with self.assertRaises(WrongSubmodel):
            self.submodel1({"_id": ObjectId(), "field1": "value", "submodel": "wrong"}, shard_id="s1")
        with self.assertRaises(MissingSubmodel):
            self.submodel1({"_id": ObjectId(), "field1": "value"}, shard_id="s1")
        with self.assertRaises(IntegrityError):
            self.submodel1({"field1": "value", "submodel": "my_submodel"}, shard_id="s1")
        with self.assertRaises(WrongSubmodel):
            obj = self.submodel1({"field1": "value"}, shard_id="s1")
            obj.submodel = "wrong"
            obj.save()

    def test_submodel_field(self):
        obj = self.submodel1({}, shard_id="s1")
        self.assertTrue(hasattr(obj, "submodel"))
        self.assertEqual(obj.submodel, self.submodel1.SUBMODEL)
        obj.save()
        obj.reload()
        self.assertEqual(obj.submodel, self.submodel1.SUBMODEL)
        db_obj = self.submodel1.get("s1", obj._id)
        self.assertEqual(db_obj.submodel, self.submodel1.SUBMODEL)

    def test_inheritance(self):

        class Submodel1(self.base_model):
            SUBMODEL = "submodel1"

        class Submodel1_1(Submodel1):
            pass

        self.assertEqual(self.base_model.collection, Submodel1.collection)
        self.assertEqual(Submodel1.collection, Submodel1_1.collection)
        self.assertEqual(Submodel1.SUBMODEL, Submodel1_1.SUBMODEL)

    def test_abstract(self):
        with self.assertRaises(IntegrityError):
            self.base_model({}, shard_id="s1")

        with self.assertRaises(IntegrityError):
            class C(self.base_model):
                pass  # no SUBMODEL
            C({}, shard_id="s1")

        with self.assertRaises(IntegrityError):
            class C(self.submodel1):
                SUBMODEL = "c"
            self.submodel1.register_submodel("c", C)

    def _create_objs(self):
        """Returns two lists of objects. Objects in the same positions only differ in their submodel"""
        values = [1, 2, 3]
        objs1 = [self.submodel1({"field1": v, "field2": v}, shard_id="s1") for v in values]
        objs2 = [self.submodel2({"field1": v, "field2": v}, shard_id="s1") for v in values]
        for obj in itertools.chain(objs1, objs2):
            obj.save()
        return objs1, objs2

    def test_isolation_find(self):
        objs1, objs2 = self._create_objs()
        self.assertCountEqual(
            self.submodel1.find("s1").all(),
            objs1,
        )
        self.assertCountEqual(
            self.submodel2.find("s1").all(),
            objs2,
        )
        self.assertCountEqual(
            self.base_model.find("s1").all(),
            objs1 + objs2,
        )

        self.assertCountEqual(
            self.submodel1.find("s1", {"field1": objs1[0].field1}).all(),
            [objs1[0]],
        )
        self.assertCountEqual(
            self.base_model.find("s1", {"field1": objs1[0].field1}).all(),
            [objs1[0], objs2[0]],
        )

    def test_isolation_update(self):
        objs1, objs2 = self._create_objs()
        obj1 = objs2[0]
        obj1.field2 = "new_value"
        self.submodel2.update_many(
            "s1",
            {"field1": obj1.field1},
            {"$set": {"field2": obj1.field2}}
        )
        self.assertCountEqual(
            self.base_model.find("s1", {"field2": obj1.field2}),
            [obj1]
        )

        obj1 = objs1[1]
        obj2 = objs2[1]
        obj1.field2 = "newer_value"
        obj2.field2 = "newer_value"
        self.base_model.update_many(
            "s1",
            {"field1": obj1.field1},
            {"$set": {"field2": obj1.field2}}
        )
        self.assertCountEqual(
            self.base_model.find("s1", {"field2": obj1.field2}),
            [obj1, obj2]
        )

    def test_isolation_destroy(self):
        objs1, objs2 = self._create_objs()
        to_destroy = objs2.pop()
        to_keep = objs1[-1]
        self.submodel2.destroy_many("s1", {"field1": to_destroy.field1})
        self.assertListEqual(
            self.submodel2.find("s1", {
                "field1": to_destroy.field1
            }).all(),
            []
        )
        self.assertListEqual(
            self.base_model.find("s1", {
                "field1": to_destroy.field1
            }).all(),
            [to_keep]
        )

        to_destroy = objs1[0]
        objs1 = objs1[1:]
        objs2 = objs2[1:]
        self.base_model.destroy_many("s1", {"field1": to_destroy.field1})
        self.assertCountEqual(
            self.base_model.find("s1").all(),
            objs1 + objs2,
        )

    def test_double_register(self):
        with self.assertRaises(IntegrityError):
            class NewSubmodel(self.base_model):
                SUBMODEL = self.submodel2.SUBMODEL
            self.base_model.register_submodel(
                NewSubmodel.SUBMODEL, NewSubmodel)




class UORMException(Exception):
    pass


class AbortTransaction(UORMException):
    def __init__(self):
        super(AbortTransaction, self).__init__("transaction aborted")


class InvalidShardId(UORMException):
    def __init__(self, shard_id):
        super(InvalidShardId, self).__init__(f"invalid shard_id \"{shard_id}\"")


class ConfigurationError(UORMException):
    pass


class ValidationError(UORMException):
    pass


class DoNotSave(UORMException):
    pass


class ObjectSaveRequired(UORMException):
    pass


class ModelDestroyed(UORMException):
    pass


class MissingShardId(UORMException):
    pass


class IntegrityError(UORMException):
    pass


class MissingSubmodel(UORMException):
    pass


class WrongSubmodel(UORMException):
    pass


class UnknownSubmodel(UORMException):
    pass



import functools

from datetime import datetime
from hashlib import md5
from flask import g, has_app_context
from random import randint
from . import ctx
from .uorm.db import ObjectsCursor


DEFAULT_CACHE_PREFIX = 'uengine'
DEFAULT_CACHE_TIMEOUT = 3600


class RequestLocalCache:

    @staticmethod
    def has(key):
        if not has_app_context():
            return False
        return key in g.request_local_cache

    @staticmethod
    def set(key, value, **kwargs):
        if not has_app_context():
            return False
        g.request_local_cache[key] = value

    @staticmethod
    def get(key):
        if not has_app_context():
            return None
        return g.request_local_cache[key]

    def delete(self, key):
        if not self.has(key):
            return False
        del g.request_local_cache[key]
        return True


def _generate_function_cache_key(pref, funcname, args, kwargs):
    key = "%s:%s(%s.%s)" % (
        pref,
        funcname,
        md5(str(args).encode("utf-8")).hexdigest(),
        md5(str(kwargs).encode("utf-8")).hexdigest()
    )

    kwargs_str = ", ".join(["%s=%s" % (x[0], x[1]) for x in kwargs.items()])
    arguments = ""
    if args:
        arguments = ", ".join([str(x) for x in args])
        if kwargs:
            arguments += ", " + kwargs_str
    else:
        if kwargs:
            arguments = kwargs
    cached_call = "%s:%s(%s)" % (pref, funcname, arguments)
    return key, cached_call


def cached_function(cache_key_prefix=DEFAULT_CACHE_PREFIX, cache_timeout=DEFAULT_CACHE_TIMEOUT, positive_only=False):
    def cache_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            cache_key, _ = _generate_function_cache_key(
                cache_key_prefix, func.__name__, args, kwargs)
            t1 = datetime.now()

            if ctx.cache.has(cache_key):
                value = ctx.cache.get(cache_key)
                ctx.log.debug("Cache HIT %s (%.3f seconds)",
                              cache_key, (datetime.now() - t1).total_seconds())
            else:
                value = func(*args, **kwargs)
                if value or not positive_only:
                    ctx.cache.set(cache_key, value, timeout=cache_timeout)
                ctx.log.debug("Cache MISS %s (%.3f seconds)",
                              cache_key, (datetime.now() - t1).total_seconds())
            return value
        return wrapper
    return cache_decorator


def cached_method(prefix, key_field=None, cache_timeout=None, positive_only=False):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            t1 = datetime.now()
            obj = args[0]
            if key_field is not None:
                if not hasattr(obj, key_field):
                    ctx.log.error(
                        f"MethodCache ERROR {obj.__class__.__name__} doesn't have attr {key_field}")
                    return func(*args, **kwargs)
                try:
                    key = str(getattr(obj, key_field))
                except TypeError as e:
                    ctx.log.error(
                        f"MethodCache ERROR parsing {key_field}: {e}")
                    return func(*args, **kwargs)

                cache_key = f"{prefix}.{key}"
            else:
                cache_key = prefix

            if ctx.cache.has(cache_key):
                value = ctx.cache.get(cache_key)
                ctx.log.debug("MethodCache HIT %s (%.3f seconds)",
                              cache_key, (datetime.now() - t1).total_seconds())
            else:
                value = func(*args, **kwargs)
                if value or not positive_only:
                    ctx.cache.set(cache_key, value, timeout=cache_timeout)
                ctx.log.debug("MethodCache MISS %s (%.3f seconds)",
                              cache_key, (datetime.now() - t1).total_seconds())

            return value

        return wrapper
    return decorator


def check_cache():
    k = md5(str(randint(0, 1000000)).encode('utf-8')).hexdigest()
    v = md5(str(randint(0, 1000000)).encode('utf-8')).hexdigest()
    ctx.cache.set(k, v)
    if ctx.cache.get(k) != v:
        return False

    ctx.cache.delete(k)
    return True


def once_per_request(cache_key_prefix=DEFAULT_CACHE_PREFIX + ".once"):
    """
    Decorator used for ensuring a subroutine runs exactly one time per api request.

    This one acts exactly like request_time_cache without returning value and checking if
    it's positive and things like that.
    """
    def cache_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not has_app_context():
                func(*args, **kwargs)
                return
            flag_key, _ = _generate_function_cache_key(
                cache_key_prefix, func.__name__, args, kwargs)
            t1 = datetime.now()

            if not req_cache_has_key(flag_key):
                func(*args, **kwargs)
                req_cache_set(flag_key, True)
                ts = (datetime.now() - t1).total_seconds()
                ctx.log.debug("OncePerRequest MISS %s(%s) (%.3f secs)", func.__name__,
                              flag_key, ts)
            else:
                ts = (datetime.now() - t1).total_seconds()
                ctx.log.debug("OncePerRequest HIT  %s(%s) (%.3f secs)", func.__name__,
                              flag_key, ts)
        return wrapper
    return cache_decorator


def request_time_cache(cache_key_prefix=DEFAULT_CACHE_PREFIX):
    """
    Decorator used for caching data during one api request.
    It's useful while some "list something" handlers with a number of cross-references generate
    many repeating database requests which are known to generate the same response during the api request.
    I.e. list of 20 hosts included in the same group and inheriting the same set of tags/custom fields
    may produce 20 additional db requests and 20 requests for each parent group recursively. This may be fixed
    by caching db responses in flask "g" store.
    """
    def cache_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not has_app_context():
                return func(*args, **kwargs)

            cache_key, _ = _get_cache_key(
                cache_key_prefix, func.__name__, args, kwargs)
            t1 = datetime.now()

            if not req_cache_has_key(cache_key):
                value = func(*args, **kwargs)
                req_cache_set(cache_key, value)
                ts = (datetime.now() - t1).total_seconds()
                ctx.log.debug("RTCache MISS %s(%s) (%.3f secs)",
                              func.__name__, cache_key, ts)
            else:
                value = req_cache_get(cache_key)
                if isinstance(value, ObjectsCursor):
                    value.cursor.rewind()
                ts = (datetime.now() - t1).total_seconds()
                ctx.log.debug("RTCache HIT  %s(%s) (%.3f secs)",
                              func.__name__, cache_key, ts)
            return value
        return wrapper
    return cache_decorator


def patch_delete_many(self, *keys):
    return all([self.delete(key) for key in keys])



import socket
import random
from pymongo import ASCENDING
from datetime import timedelta
from time import time, sleep

from .base_queue import BaseQueue, DEFAULT_ACK_TIMEOUT, DEFAULT_RETRIES
from .task import BaseTask, TaskSendError

from ..utils import now
from ..context import ctx


class MongoQueue(BaseQueue):

    def __init__(self, qcfq):
        super(MongoQueue, self).__init__(qcfq)
        self._initialized = False
        self.task_collection = self.cfg.get("collection", "mq_tasks")
        self.subs_collection = self.task_collection + "_subs"
        self.prefix = self.cfg.get("channel", "ueq")
        self.channel_ttl = self.cfg.get("channel_ttl", 5)
        self.ack_timeout = self.cfg.get("ack_timeout", DEFAULT_ACK_TIMEOUT)
        self.retries = self.cfg.get("retries", DEFAULT_RETRIES)

        fqdn = socket.gethostname()
        rand = random.randint(0, 10000)
        self.msgchannel = f"{self.prefix}:{fqdn}:{rand}"
        self.ackchannel = f"{self.prefix}:{fqdn}:{rand}:ack"

    def initialize(self):
        self.ensure_indexes()
        self.cleanup_channels()
        self._initialized = True

    def ensure_indexes(self):
        self.coll_subs.ensure_index("updated_at")
        self.coll_subs.ensure_index("chan")
        self.coll_tasks.ensure_index(
            [("chan", ASCENDING), ("created_at", ASCENDING)])

    def cleanup_channels(self):
        min_date = now() - timedelta(seconds=self.channel_ttl)
        self.coll_subs.delete_many({"updated_at": {"$lt": min_date}})

    @property
    def coll_subs(self):
        return ctx.db.meta.conn[self.subs_collection]

    @property
    def coll_tasks(self):
        return ctx.db.meta.conn[self.task_collection]

    def subscribe(self):
        self.coll_subs.replace_one({"chan": self.msgchannel},
                                   {"chan": self.msgchannel, "updated_at": now()}, upsert=True)

    def list_active_channels(self):
        min_date = now() - timedelta(seconds=self.channel_ttl)
        channels = list(self.coll_subs.find({"updated_at": {"$gt": min_date}}))
        return channels

    def get_random_channel(self):
        channels = self.list_active_channels()
        if len(channels) == 0:
            return None, None
        rand = random.randrange(0, len(channels))
        chan_name = channels[rand]["chan"]
        ack_chan_name = chan_name + self.ACK_POSTFIX
        return chan_name, ack_chan_name

    def wait_ack(self, ins_id, chan):
        cancel_at = time() + self.ack_timeout
        query = {"ins_id": ins_id, "chan": chan}
        while time() < cancel_at:
            c = self.coll_tasks.find_one(query)
            if c:
                self.coll_tasks.delete_one(query)
                return c
            sleep(0.01)
        return None

    def ack(self, task_id):
        # generate ack
        ctx.log.debug("generating ACK {ins_id: %s, chan: %s}", task_id, self.ackchannel)
        self.coll_tasks.insert_one(
            {"ins_id": task_id, "chan": self.ackchannel})
        # remove task doc
        ctx.log.debug("removing task doc from collection {_id: %s}", task_id)
        self.coll_tasks.delete_one({"_id": task_id})

    def publish(self, chan, message):
        res = self.coll_tasks.insert_one(
            {"chan": chan, "data": message, "created_at": now()})
        return res.inserted_id

    def _enqueue(self, task):
        if not self._initialized:
            self.initialize()

        ack = None
        retries = self.retries
        while retries > 0:
            chan, ackchan = self.get_random_channel()
            if chan is None:
                sleep(.5)
                retries -= 1
                if retries > 0:
                    ctx.log.error(
                        "no active channels found, resending, %d retries left", retries)
                    continue
                else:
                    raise TaskSendError("no active channels")
            ins_id = self.publish(chan, task.to_message())
            ack = self.wait_ack(ins_id, ackchan)
            if ack:
                break
            retries -= 1
            if retries > 0:
                ctx.log.error("error receiving ack for task id %s, resending, %d retries left",
                              task.id, retries)
            else:
                raise TaskSendError(
                    f"error receiving ack after {self.retries} retries")

        recvchan = ack["chan"]
        receiver = recvchan[len(self.prefix) + 1:-len(self.ACK_POSTFIX)]
        task.set_recv_by(receiver)
        return ack

    @property
    def tasks(self):
        resub_interval = timedelta(seconds=self.channel_ttl//2)
        if not self._initialized:
            self.initialize()
        self.subscribe()
        resub_at = now() + resub_interval
        while True:
            items = self.coll_tasks.find(
                {"chan": self.msgchannel}).sort("created_at", ASCENDING)

            cnt = items.count()
            if cnt > 0:
                ctx.log.debug("processing %d tasks from task collection", cnt)
                for item in items:
                    task = BaseTask.from_message(item["data"])
                    ctx.log.debug("task %s created from item %s", task.id, item)
                    ctx.log.debug("setting task %s received by", task.id)
                    task.set_recv_by(self.msgchannel[len(self.prefix) + 1:])
                    ctx.log.debug("sending ack for task %s", task.id)
                    self.ack(item["_id"])
                    yield task
            else:
                sleep(0.1)
            if now() > resub_at:
                self.subscribe()
                resub_at = now() + resub_interval



import os
import sys
import inspect
import logging

from flask import Flask, g, request
from datetime import timedelta
from logging.handlers import WatchedFileHandler
from cachelib import MemcachedCache, SimpleCache
from uuid import uuid4

from . import ctx
from .errors import handle_api_error, handle_other_errors, ApiError
from .sessions import MongoSessionInterface
from .json_encoder import MongoJSONEncoder
from .queue import RedisQueue, MongoQueue, DummyQueue
from .uorm.db import db

ENVIRONMENT_TYPES = ("development", "testing", "production")
DEFAULT_ENVIRONMENT_TYPE = "development"
DEFAULT_SESSION_EXPIRATION_TIME = 86400 * 7 * 2
DEFAULT_TOKEN_TTL = 86400 * 7 * 2
DEFAULT_LOG_FORMAT = "[%(asctime)s] %(levelname)s %(filename)s:%(lineno)d %(request_id)s %(message)s"
DEFAULT_LOG_LEVEL = "debug"
DEFAULT_FILECACHE_DIR = "/var/cache/uengine"


class RequestIDFilter(logging.Filter):
    def filter(self, record):
        if request:
            if 'X-Request-ID' in request.headers:
                record.request_id = request.headers.get('X-Request-ID')
            elif 'request_id' in request.args:
                record.request_id = request.args.get('request_id')
            else:
                record.request_id = getattr(g, "request_id", None)
        else:
            record.request_id = None

        return True


class Base:

    def __init__(self):
        class_file = inspect.getfile(self.__class__)
        self.app_dir = os.path.dirname(os.path.abspath(class_file))
        self.base_dir = os.path.abspath(os.path.join(self.app_dir, "../"))
        self.app_name = os.path.basename(self.app_dir)

        self.version = "development"
        self.__set_version()

        envtype = os.getenv("GLASS_ENV", DEFAULT_ENVIRONMENT_TYPE)
        if envtype not in ENVIRONMENT_TYPES:
            envtype = DEFAULT_ENVIRONMENT_TYPE
        ctx.envtype = envtype

        self.session_expiration_time = None
        self.session_auto_cleanup = None
        self.session_auto_cleanup_trigger = None

        # Later steps depend on the earlier ones. The order is important here
        ctx.cfg = self.__read_config()
        ctx.log = self.__setup_logging()

        ctx.cache = self.__setup_cache()
        ctx.req_cache = self.__setup_request_cache()

        ctx.db = self.__setup_db()  # Requires ctx.cfg
        ctx.queue = self.__setup_queue()
        self.flask = self.__setup_flask()  # requires ctx.cfg and ctx.log
        self.__setup_error_handling()  # requires self.flask
        self.__setup_sessions()
        self.setup_controllers()
        self.after_setup()

    def setup_controllers(self):
        pass

    def after_setup(self):
        pass

    def __set_version(self):
        ver_filename = os.path.join(self.base_dir, "__version__")
        if not os.path.isfile(ver_filename):
            return
        with open(ver_filename) as verf:
            self.version = verf.read().strip()

    @staticmethod
    def __setup_db():
        db.configure(ctx.cfg.get("database"))
        db.setup_cache(ctx.req_cache, 1)
        db.setup_cache(ctx.cache, 2)
        return db

    @staticmethod
    def __setup_cache():
        ctx.log.debug("Setting up a cache")
        if "memcache_backends" in ctx.cfg:
            return MemcachedCache(ctx.cfg.get("memcache_backends"))

        from .cache import patch_delete_many
        SimpleCache.delete_many = patch_delete_many
        return SimpleCache()

    @staticmethod
    def __setup_request_cache():
        ctx.log.debug("Setting up a request local cache")
        from .cache import RequestLocalCache
        return RequestLocalCache()

    @staticmethod
    def __setup_queue():
        ctx.log.debug("Setting up a queue")
        qcfg = ctx.cfg.get("queue", {})
        qtype = qcfg.get("type", "mongo")  # default internal naive queue

        if qtype == "redis":
            try:
                q = RedisQueue(qcfg)
                return q
            except Exception as e:
                ctx.log.error("Error configuring redis queue: %s", e)
        elif qtype == "mongo":
            try:
                q = MongoQueue(qcfg)
                return q
            except Exception as e:
                ctx.log.error("Error configuring mongo queue: %s", e)
        elif qtype == "dummy":
            try:
                q = DummyQueue(qcfg)
                return q
            except Exception as e:
                ctx.log.error("Error configuring dummy queue: %s", e)

    @staticmethod
    def __setup_logging():
        logger = logging.getLogger("app")
        logger.propagate = False

        log_level = ctx.cfg.get("log_level", DEFAULT_LOG_LEVEL)
        log_level = log_level.upper()
        log_level = getattr(logging, log_level)

        if "log_file" in ctx.cfg:
            handler = WatchedFileHandler(ctx.cfg.get("log_file"))
            logger.addHandler(handler)

        if ctx.cfg.get("debug") or not logger.handlers:
            handler = logging.StreamHandler(stream=sys.stdout)
            logger.addHandler(handler)

        log_format = ctx.cfg.get("log_format", DEFAULT_LOG_FORMAT)
        log_format = logging.Formatter(log_format)

        logger.setLevel(log_level)
        for handler in logger.handlers:
            handler.setLevel(log_level)
            handler.setFormatter(log_format)

        logger.addFilter(RequestIDFilter())

        logger.info("Logger created, starting up")
        return logger

    @staticmethod
    def __setup_flask():
        ctx.log.info("Environment type is %s", ctx.envtype)
        ctx.log.debug("Setting up Flask")
        flask = Flask(__name__, static_folder=None)

        ctx.log.debug("Applying Flask settings")
        if "flask_settings" in ctx.cfg:
            for k, v in ctx.cfg.get("flask_settings").items():
                flask.config[k] = v
        flask.secret_key = ctx.cfg.get("app_secret_key")
        ctx.log.debug("Setting up JSON encoder")
        flask.json_encoder = MongoJSONEncoder

        # pylint: disable=unused-variable
        @flask.before_request
        def add_request_local_cache():
            g.request_local_cache = {}

        if ctx.cfg.get("debug"):
            ctx.log.info("Setting up request logging due to debug setting")

            try:
                from . import profilers
                flask.before_request(profilers.before_request)
                flask.after_request(profilers.after_request)
            except ImportError:
                ctx.log.exception("error importing profiler, profiling will be disabled")

            # pylint: disable=unused-variable
            @flask.before_request
            def add_request_id():
                setattr(g, "request_id", str(uuid4())[:8])

            @flask.before_request
            def log_all_requests():
                msg = "REQ_%s %s data=%s" % (
                    request.method, request.path, request.json)
                ctx.log.debug(msg)

            @flask.after_request
            def log_all_responses(response):
                if response.content_type == 'application/json':
                    ctx.log.debug(" ".join(map(str, [
                        response.status,
                        str(response.headers).rstrip('\r\n'),
                        response.get_data(),
                    ])))
                else:
                    ctx.log.debug(
                        " ".join(map(str, [response.status, str(response.headers).rstrip('\r\n')])))
                return response

        return flask

    def __read_config(self):
        config_filename = os.path.join(
            self.base_dir, "config", "%s.py" % ctx.envtype)
        with open(config_filename) as f:
            config = {}
            text = f.read()
            code = compile(text, config_filename, 'exec')
            exec(code, config)  # pylint: disable=exec-used
            del config["__builtins__"]
            return config

    def __setup_sessions(self):
        ctx.log.debug("Setting up sessions")

        e_time = ctx.cfg.get("session_ttl", DEFAULT_SESSION_EXPIRATION_TIME)
        self.flask.session_interface = MongoSessionInterface(
            collection_name='sessions')
        self.flask.permanent_session_lifetime = timedelta(seconds=e_time)
        self.session_expiration_time = timedelta(seconds=e_time)
        self.session_auto_cleanup = ctx.cfg.get("session_auto_cleanup", True)
        self.session_auto_cleanup_trigger = ctx.cfg.get(
            "session_auto_cleanup_rand_trigger", 0.05)

    def __setup_error_handling(self):
        self.flask.register_error_handler(ApiError, handle_api_error)
        self.flask.register_error_handler(Exception, handle_other_errors)

    def run(self, **kwargs):
        self.flask.run(**kwargs)



import re
from bson import ObjectId
from datetime import datetime
from glasskit.uorm.errors import ValidationError
from pymongo import ASCENDING


class BaseFieldDescriptor:
    def __set__(self, instance, value):
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name


class AutotrimFieldDescriptor(BaseFieldDescriptor):
    def __set__(self, instance, value):
        if hasattr(value, "strip"):
            value = value.strip()
        super().__set__(instance, value)


class Field:

    descriptor = BaseFieldDescriptor

    def __init__(
        self,
        required=False,
        rejected=False,
        restricted=False,
        default=None,
        index=None,
        unique=False,
        choices=None,
    ):
        self.required = required
        self.rejected = rejected
        self.restricted = restricted
        if choices:
            self.choices = set(choices)
        else:
            self.choices = None
        self.def_value = default
        self.index, self.index_options = self.__generate_index(index, unique)
        self.name = None

    @staticmethod
    def __generate_index(index, unique):
        if unique:
            if index is None or index is True:
                return ASCENDING, {"unique": True}
            else:
                return index, {"unique": True}
        else:
            if index is None:
                return None, {}
            elif index is True:
                return ASCENDING, {}
            else:
                return index, {}

    def set_name(self, name):
        self.name = name

    def validate(self, value):
        if self.required and value is None:
            raise ValidationError(f"field {self.name} is required")

        if self.choices and value not in self.choices:
            if value is None:
                # at this point the value can be None only if it's not required
                # thus, if it's not required, it's ok to be None even if choices are defined
                return

            raise ValidationError(f"field {self.name} must be one of {self.choices}")


class TypedField(Field):

    TYPE = None

    def validate(self, value):
        super().validate(value)
        if value is None and not self.required:
            return
        if not isinstance(value, self.TYPE):
            raise ValidationError(
                f"field {self.name} must be an instance of type {self.TYPE.__name__}"
            )


class ObjectIdField(TypedField):

    TYPE = ObjectId


class StringField(TypedField):

    TYPE = str

    def __init__(
        self, min_length=None, max_length=None, re_match=None, auto_trim=True, **kwargs
    ):
        super().__init__(**kwargs)
        if auto_trim:
            self.__class__.descriptor = AutotrimFieldDescriptor

        if min_length is not None:
            assert isinstance(min_length, int)
        if max_length is not None:
            assert isinstance(max_length, int)

        self.min_length = min_length
        self.max_length = max_length
        if re_match:
            self.re_match = re.compile(re_match)
        else:
            self.re_match = None

    def validate(self, value):
        super().validate(value)
        if self.min_length is not None and len(value) < self.min_length:
            raise ValidationError(
                f"field {self.name} must be at least {self.min_length} characters long"
            )

        if self.max_length is not None and len(value) > self.max_length:
            raise ValidationError(
                f"field {self.name} must be at most {self.max_length} characters long"
            )

        if self.re_match is not None and not self.re_match.match(value):
            raise ValidationError(
                f'field {self.name} must match pattern "{self.re_match.pattern}"'
            )


class IntField(TypedField):

    TYPE = int

    def __init__(self, min_value=None, max_value=None, **kwargs):
        super().__init__(**kwargs)
        if min_value is not None:
            assert isinstance(min_value, int)
        if max_value is not None:
            assert isinstance(max_value, int)
        self.min_value = min_value
        self.max_value = max_value

    def validate(self, value):
        super().validate(value)
        if self.min_value is not None and value < self.min_value:
            raise ValidationError(f"field {self.name} must be >= {self.min_value}")
        if self.max_value is not None and value > self.max_value:
            raise ValidationError(f"field {self.name} must be <= {self.max_value}")


class FloatField(TypedField):

    TYPE = float

    def __init__(self, min_value=None, max_value=None, **kwargs):
        super().__init__(**kwargs)
        if min_value is not None:
            assert isinstance(min_value, float) or isinstance(min_value, int)
        if max_value is not None:
            assert isinstance(max_value, float) or isinstance(max_value, int)
        self.min_value = min_value
        self.max_value = max_value

    def validate(self, value):
        super().validate(value)
        if self.min_value is not None and value < self.min_value:
            raise ValidationError(f"field {self.name} must be >= {self.min_value}")
        if self.max_value is not None and value > self.max_value:
            raise ValidationError(f"field {self.name} must be <= {self.max_value}")


class ListField(TypedField):

    TYPE = list

    def __init__(self, min_length=None, max_length=None, **kwargs):
        super().__init__(**kwargs)
        if min_length is not None:
            assert isinstance(min_length, int)
        if max_length is not None:
            assert isinstance(max_length, int)
        self.min_length = min_length
        self.max_length = max_length

    def validate(self, value):
        super().validate(value)
        if self.min_length is not None and len(value) < self.min_length:
            raise ValidationError(
                f"field {self.name} must be at least {self.min_length} items long"
            )

        if self.max_length is not None and len(value) > self.max_length:
            raise ValidationError(
                f"field {self.name} must be at most {self.max_length} items long"
            )


class DatetimeField(TypedField):

    TYPE = datetime


class DictField(TypedField):

    TYPE = dict


class BoolField(TypedField):

    TYPE = bool



from unittest import TestCase
from datetime import datetime
from ..uorm.models.fields import (
    StringField,
    IntField,
    FloatField,
    ListField,
    DictField,
    BoolField,
    DatetimeField,
)
from ..uorm.models.base_model import BaseModel
from ..uorm.errors import ValidationError


class TestValidators(TestCase):
    def test_required(self):
        class Model(BaseModel):
            field: StringField(required=True)

        model = Model()
        self.assertRaises(ValidationError, model.validate)
        model.field = "string"
        model.validate()

    def test_string_field(self):
        class Model(BaseModel):
            field: StringField()

        model = Model({"field": 1})
        self.assertRaises(ValidationError, model.validate)
        model.field = "string"
        model.validate()

        model.field = "   auto-trimmed   "
        self.assertEqual(model.field, "auto-trimmed")

        class Model(BaseModel):
            field: StringField(auto_trim=False)

        model = Model({"field": "    no-auto-trim    "})
        self.assertEqual(model.field, "    no-auto-trim    ")

        class Model(BaseModel):
            field: StringField(min_length=3, max_length=5, re_match="^\d+$")

        model = Model({"field": "1"})
        self.assertRaises(ValidationError, model.validate)
        model.field = "123456"
        self.assertRaises(ValidationError, model.validate)
        model.field = "1234a"
        self.assertRaises(ValidationError, model.validate)
        model.field = "1234"
        model.validate()

    def test_list_field(self):
        class Model(BaseModel):
            field: ListField()

        model = Model({"field": 1})
        self.assertRaises(ValidationError, model.validate)
        model.field = [1, 2, 3]
        model.validate()

        class Model(BaseModel):
            field: ListField(min_length=3, max_length=5)

        model = Model({"field": [1]})
        self.assertRaises(ValidationError, model.validate)
        model.field = [1, 2, 3, 4, 5, 6]
        self.assertRaises(ValidationError, model.validate)
        model.field = [1, 2, 3, 4]
        model.validate()

    def test_bool_field(self):
        class Model(BaseModel):
            field: BoolField()

        model = Model({"field": 1})
        self.assertRaises(ValidationError, model.validate)
        model.field = False
        model.validate()

    def test_int_field(self):
        class Model(BaseModel):
            field: IntField()

        model = Model({"field": "string"})
        self.assertRaises(ValidationError, model.validate)
        model.field = -34
        model.validate()

        class Model(BaseModel):
            field: IntField(min_value=0, max_value=10)

        model = Model({"field": -34})
        self.assertRaises(ValidationError, model.validate)
        model.field = 34
        self.assertRaises(ValidationError, model.validate)
        model.field = 5
        model.validate()

    def test_float_field(self):
        class Model(BaseModel):
            field: FloatField()

        model = Model({"field": "string"})
        self.assertRaises(ValidationError, model.validate)
        model.field = -34
        self.assertRaises(ValidationError, model.validate)
        model.field = 28.0
        model.validate()

        class Model(BaseModel):
            field: FloatField(min_value=0, max_value=10)

        model = Model({"field": -15.5})
        self.assertRaises(ValidationError, model.validate)
        model.field = 15.5
        self.assertRaises(ValidationError, model.validate)
        model.field = 5.3
        model.validate()

    def test_dict_field(self):
        class Model(BaseModel):
            field: DictField()

        model = Model({"field": 1})
        self.assertRaises(ValidationError, model.validate)
        model.field = {}
        model.validate()

    def test_datetime_field(self):
        class Model(BaseModel):
            field: DatetimeField()

        model = Model({"field": 1})
        self.assertRaises(ValidationError, model.validate)
        model.field = datetime.now()
        model.validate()

    def test_choices(self):
        class Model(BaseModel):
            field: StringField(required=True, choices=["a", "b", "c"])

        class ModelNotRequired(BaseModel):
            field: StringField(required=False, choices=["a", "b", "c"])

        model = Model({"field": "a"})
        model.validate()
        model.field = "b"
        model.validate()
        model.field = "c"
        model.validate()
        model.field = "d"
        self.assertRaises(ValidationError, model.validate)
        model.field = None
        self.assertRaises(ValidationError, model.validate)

        model = ModelNotRequired({"field": "a"})
        model.validate()
        model.field = "b"
        model.validate()
        model.field = "c"
        model.validate()
        model.field = "d"
        self.assertRaises(ValidationError, model.validate)
        model.field = None
        model.validate()



import os
import sys
import inspect
import importlib.util
from argparse import ArgumentParser, REMAINDER


class Command:
    """Base class for CLI commands
    You should override at least run() method
        run() - method is being called to do the main job
        init_argument_parser(parser) - override this method if you want to accept arguments
        NAME - if this property is not None it is used as command name.
               Otherwise command name is generated from class name
        DESCRIPTION - description that is used in help messages. Consider setting it to something meaningful.
    """
    NAME = None
    DESCRIPTION = None
    NO_ARGPARSE = False

    def __init_subclass__(cls, **kwargs):
        if cls.NAME is None:
            cls.NAME = cls.__name__.lower()

    def __init__(self, app):
        self.app = app
        if self.NAME is None:
            self.NAME = self.__class__.__name__.lower()
        if self.DESCRIPTION is None:
            self.DESCRIPTION = '"%s" has no DESCRIPTION' % (self.NAME,)
        self.args = None
        self.raw_args = None

    def init_argument_parser(self, parser):
        """
        This method is called to configure argument subparser for command
        Override it if you need to accept arguments
          - parser: argparse.ArgumentParser to fill with arguments
        """

    def run(self):
        """
        This is a main method of command. Override it and do all the job here
        Command arguments can be read from self.args
        The return value from this method will be used as CLI exit code
        """
        raise NotImplementedError()


def is_a_command_class(obj):
    return inspect.isclass(obj) and issubclass(obj, Command) and obj != Command


def load_commands():
    commands = {}
    for mod_pref in ["glasskit.commands", "commands"]:
        commands_package = importlib.util.find_spec(mod_pref).origin
        for pyfile in os.listdir(os.path.dirname(commands_package)):
            if pyfile.endswith(".py"):
                module_name = f"{mod_pref}.{pyfile[:-3]}"
                module = importlib.import_module(module_name)
                for cls_name in dir(module):
                    cls = getattr(module, cls_name)
                    if is_a_command_class(cls):
                        commands[cls.NAME] = cls
    return commands


def main(app):

    parser = ArgumentParser()
    subparsers = parser.add_subparsers(
        title='Commands',
        help="One of the following commands",
        description='use <command> --help to get help on particular command',
        metavar="<command>",
    )

    for cmd_name, cmd_class in load_commands().items():
        command = cmd_class(app)
        if command.NO_ARGPARSE:
            command_parser = subparsers.add_parser(
                command.NAME,
                help=command.DESCRIPTION,
                add_help=False,
                # Ugly hack to prevent arguments from being parsed as options
                prefix_chars=chr(0),
            )
            command_parser.add_argument('raw_args', nargs=REMAINDER)
        else:
            command_parser = subparsers.add_parser(
                command.NAME, help=command.DESCRIPTION)
        command_parser.set_defaults(command=command)
        command.init_argument_parser(command_parser)

    args = parser.parse_args()
    if not hasattr(args, "command"):
        print(f"Usage: {sys.argv[0]} <command> [<options>]")
    else:
        args.command.args = args
        if 'raw_args' in args:
            args.command.raw_args = args.raw_args
        return args.command.run()



from threading import Thread
from queue import Queue, Empty
from glasskit import ctx
from time import sleep


class BaseWorker(Thread):

    def __init__(self, empty_queue_sleep=0.1, task_retries=5, between_retries_sleep=5):
        super(BaseWorker, self).__init__()
        self.stopped = False
        self.eq_sleep = empty_queue_sleep
        self.retries = task_retries
        self.br_sleep = between_retries_sleep

    def run_task(self, task):
        """
        override this method
        """
        raise NotImplementedError("abstract worker")

    def run(self):
        ctx.log.debug("task-processing worker started")
        while not self.stopped:
            try:
                task = self.q.get(False)
                ctx.log.debug("Got task %s", task)
            except Empty:
                sleep(self.eq_sleep)
                continue
            task_retries = self.retries
            done = False
            while task_retries > 0:
                try:
                    self.run_task(task)
                    done = True
                    break
                except Exception as e:
                    ctx.log.error("error executing task %s: %s", task.id, e)
                    task_retries -= 1
                    if task_retries > 0:
                        ctx.log.info("about to restart task %s in %d seconds, retries left %d",
                                     task.id, self.br_sleep, task_retries)
                    sleep(self.br_sleep)
            if not done:
                ctx.log.error("task %s failed, giving up", task.id)

    def process_tasks(self):
        self.q = Queue()
        self.start()

        try:
            while True:
                try:
                    for task in ctx.queue.tasks:
                        self.q.put(task)
                except NotImplementedError:
                    ctx.log.error("you are probably using BaseWorker itself while it's an abstract class")
                    break
                except Exception as e:
                    ctx.log.error("error moving task from task queue to worker: %s", e)
        except KeyboardInterrupt:
            self.stopped = True
            self.join()


from flask import Blueprint, g
from .errors import AuthenticationError


class Controller(Blueprint):
    def __init__(
        self, name, import_name, require_auth=True, auth_error_class=None, **kwargs
    ):
        super().__init__(name, import_name, **kwargs)

        if auth_error_class is None:
            auth_error_class = AuthenticationError
        if type(auth_error_class) != type or not issubclass(
            auth_error_class, Exception
        ):
            raise TypeError("auth_error_class must be a subclass of Exception")
        self.auth_error = auth_error_class

        self.before_request(self.__set_auth_data)
        if require_auth:
            self.before_request(self.check_auth)

    def get_current_user(self) -> object:
        return None

    def __set_auth_data(self):
        g.user = self.get_current_user()

    def check_auth(self):
        if g.user is None:
            raise self.auth_error()



from ..context import ctx
from ..utils import uuid4_string, now
from flask import json


class BaseTask:

    TYPE = "BASE"
    TYPE_MAP = {}

    def __init__(self, data, created_at=None, task_id=None):
        self.id = task_id or uuid4_string()
        self.data = data
        self.created_at = created_at or now()
        self.received_by = None

    def to_message(self):
        return {
            "id": self.id,
            "type": self.TYPE,
            "data": json.dumps(self.data),
            "created_at": self.created_at
        }

    def set_recv_by(self, recv_by):
        self.received_by = recv_by

    @property
    def received(self):
        return self.received_by is not None

    @classmethod
    def from_message(cls, msg):
        if "data" not in msg:
            ctx.log.error("malformed message, no data field: %s", msg)
        task_id = msg["id"]
        task_type = msg["type"]
        created_at = msg["created_at"]
        data = json.loads(msg["data"])
        if task_type in cls.TYPE_MAP:
            task_class = cls.TYPE_MAP[task_type]
        else:
            task_class = cls
        return task_class(task_id=task_id, data=data, created_at=created_at)

    @classmethod
    def register(cls):
        BaseTask.TYPE_MAP[cls.TYPE] = cls

    def publish(self):
        return ctx.queue.enqueue(self)

    def __str__(self):
        return f"<{self.__class__.__name__} {self.TYPE} id={self.id} data={json.dumps(self.data)} " + \
               f"created_at={self.created_at} received_by={self.received_by}>"

    def __repr__(self):
        return self.__str__()


class BasePingTask(BaseTask):
    TYPE = "BASE_PING"


BasePingTask.register()


class TaskSendError(Exception):
    pass



from ..uorm.db import db
from ..uorm.models.sharded_model import ShardedModel
from ..uorm.errors import MissingShardId
from ..uorm.models.fields import StringField, Field
from .mongo_mock_test import MongoMockTest

CALLABLE_DEFAULT_VALUE = 4


def callable_default():
    return CALLABLE_DEFAULT_VALUE


class TestModel(ShardedModel):
    field1: StringField(rejected=True, index=True, default="default_value")
    field2: Field(required=True)
    field3: Field(required=True, default="required_default_value")
    callable_default_field: Field(default=callable_default)


class TestShardedModel(MongoMockTest):

    def setUp(self) -> None:
        super().setUp()
        for shard_id in db.shards:
            TestModel.destroy_all(shard_id)

    def tearDown(self) -> None:
        for shard_id in db.shards:
            TestModel.destroy_all(shard_id)
        super().tearDown()

    def test_init(self):
        model = TestModel({"field1": "value"})
        self.assertEqual(model.field1, "value")
        model._before_save()
        model._before_delete()

    def test_shard(self):
        model = TestModel({"field2": "value"})
        self.assertIsNone(model._shard_id)
        self.assertRaises(MissingShardId, model.save)

        shard_id = db.rw_shards[0]
        model = TestModel(shard_id=shard_id, attrs={"field2": "value"})
        self.assertEqual(model._shard_id, shard_id)
        model.save()



from .fields import Field


def snake_case(name):
    result = ""
    for i, l in enumerate(name):
        if 65 <= ord(l) <= 90:
            if i != 0:
                result += "_"
            result += l.lower()
        else:
            result += l
    return result


class MetaModel:

    __fields__ = None
    __rejected_fields__ = None
    __restricted_fields__ = None
    __validators__ = None
    __def_values__ = None
    __def_indexes__ = None
    __cache_key_fields__ = None

    COLLECTION = None
    INDEXES = None
    KEY_FIELD = "_id"
    CACHE_KEY_FIELDS = None

    collection = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.__fields__ = set() if cls.__fields__ is None else cls.__fields__.copy()
        cls.__def_values__ = (
            {} if cls.__def_values__ is None else cls.__def_values__.copy()
        )
        cls.__validators__ = (
            {} if cls.__validators__ is None else cls.__validators__.copy()
        )
        cls.__rejected_fields__ = (
            set() if cls.__rejected_fields__ is None else cls.__rejected_fields__.copy()
        )
        cls.__restricted_fields__ = (
            set()
            if cls.__restricted_fields__ is None
            else cls.__restricted_fields__.copy()
        )
        cls.__def_indexes__ = (
            [] if cls.__def_indexes__ is None else cls.__def_indexes__[:]
        )
        cls.__cache_key_fields__ = (
            {"_id"}
            if cls.__cache_key_fields__ is None
            else cls.__cache_key_fields__.copy()
        )

        for name, value in cls.__annotations__.items():
            if not isinstance(value, Field):
                continue
            value.set_name(name)
            field = value.descriptor()
            field.__set_name__(cls, name)
            setattr(cls, name, field)
            cls.__fields__.add(name)
            cls.__validators__[name] = value.validate
            if value.def_value is not None:
                cls.__def_values__[name] = value.def_value
            if value.rejected:
                cls.__rejected_fields__.add(name)
            if value.restricted:
                cls.__restricted_fields__.add(name)
            if value.index is not None:
                cls.__def_indexes__.append([[(name, value.index)], value.index_options])

        if cls.INDEXES is not None:
            for idx in cls.INDEXES:
                cls.__def_indexes__.append(idx)

        cls.__cache_key_fields__.add(cls.KEY_FIELD)
        if cls.CACHE_KEY_FIELDS:
            for field in cls.CACHE_KEY_FIELDS:
                cls.__cache_key_fields__.add(field)

        if cls.COLLECTION:
            cls.collection = cls.COLLECTION
        else:
            cls.collection = snake_case(cls.__name__)
